"""Test for ContainerOp transformers."""
import pytest
import kfp.dsl

from kfx.dsl._transformers import ContainerOpTransform


@pytest.fixture
def op() -> kfp.dsl.ContainerOp:
    return kfp.dsl.ContainerOp(
        name="hello",
        image="bash",
        sidecars=[
            kfp.dsl.Sidecar(name="foo", image="bash"),
            kfp.dsl.Sidecar(name="bar", image="bash"),
        ],
    )


def test_containerop_transform_image_pull_policy(op: kfp.dsl.ContainerOp):

    transform = (
        ContainerOpTransform()
        .set_image_pull_policy("Always")
        .set_sidecar_image_pull_policy("Always", "f*")
    )

    op.apply(transform)
    assert op.container.image_pull_policy == "Always"
    assert [sidecar.image_pull_policy for sidecar in op.sidecars] == ["Always", None]


def test_containerop_transform_set_resources(op: kfp.dsl.ContainerOp):
    transform = ContainerOpTransform().set_resources(cpu=(1, 2), memory=("1G", "2G"))
    op.apply(transform)
    assert op.container.resources.requests == {"cpu": "1", "memory": "1G"}
    assert op.container.resources.limits == {"cpu": "2", "memory": "2G"}

    transform = ContainerOpTransform().set_resources(cpu="500m", memory="4G")
    op.apply(transform)
    assert op.container.resources.requests == {"cpu": "500m", "memory": "4G"}
    assert op.container.resources.limits == {"cpu": "500m", "memory": "4G"}


def test_containerop_transform_set_sidecar_resources(op: kfp.dsl.ContainerOp):

    transform = ContainerOpTransform().set_sidecar_resources(cpu="500m", memory="4G")
    op.apply(transform)
    assert [sidecar.resources.requests for sidecar in op.sidecars] == [
        {"cpu": "500m", "memory": "4G"},
        {"cpu": "500m", "memory": "4G"},
    ]
    assert [sidecar.resources.limits for sidecar in op.sidecars] == [
        {"cpu": "500m", "memory": "4G"},
        {"cpu": "500m", "memory": "4G"},
    ]

    transform = (
        ContainerOpTransform()
        .set_sidecar_resources(cpu=(1, 2), memory=("1G", "2G"), sidecar_name="f*")
        .set_sidecar_resources(cpu="200m", memory="3G", sidecar_name="bar")
    )
    op.apply(transform)
    assert [sidecar.resources.requests for sidecar in op.sidecars] == [
        {"cpu": "1", "memory": "1G"},
        {"cpu": "200m", "memory": "3G"},
    ]
    assert [sidecar.resources.limits for sidecar in op.sidecars] == [
        {"cpu": "2", "memory": "2G"},
        {"cpu": "200m", "memory": "3G"},
    ]


def test_containerop_transform_set_annotation_labels(op: kfp.dsl.ContainerOp):
    transform = (
        ContainerOpTransform()
        .set_annotations({"foo": "bar"})
        .set_labels({"hello": "world"})
    )
    op.apply(transform)

    assert op.pod_annotations == {"foo": "bar"}
    assert op.pod_labels == {"hello": "world"}


def test_containerop_transform_add_envs(op: kfp.dsl.ContainerOp):
    transform = (
        ContainerOpTransform()
        .add_env_var("foo", "bar")
        .add_env_vars({"hello": "world"})
        .add_env_var_from_secret(
            "creds", secret_name="k8s_secret", secret_key="access_key"
        )
        .add_env_var_from_configmap("some_configmap")
    )
    op.apply(transform)

    assert [obj.to_dict() for obj in op.container.env] == [
        {"name": "foo", "value": "bar", "value_from": None},
        {"name": "hello", "value": "world", "value_from": None},
        {
            "name": "creds",
            "value": None,
            "value_from": {
                "config_map_key_ref": None,
                "field_ref": None,
                "resource_field_ref": None,
                "secret_key_ref": {
                    "key": "access_key",
                    "name": "k8s_secret",
                    "optional": None,
                },
            },
        },
    ]

    assert [obj.to_dict() for obj in op.container.env_from] == [
        {
            "config_map_ref": {"name": "some_configmap", "optional": None},
            "prefix": None,
            "secret_ref": None,
        }
    ]



"""Extensions for visualizations in kubeflow.

::

    import functools

    import kfp.components


    # install kfx
    kfx_component = functools.partial(kfp.components.func_to_container_op, packages_to_install=["kfx"])


    @kfx_component
    def some_op(
        # mlpipeline_metrics is a path - i.e. open(mlpipeline_metrics, "w")
        mlpipeline_metrics: kfp.components.OutputPath(str),
        # mlpipeline_ui_metadata is a FileLike obj - i.e. mlpipeline_ui_metadata.write("something")
        mlpipeline_ui_metadata: kfp.components.OutputTextFile(str),
    ):
        "kfp operator that provides metrics and metadata for visualizations."

        # import inside kfp task
        import kfx.vis

        # output metrics to mlpipeline_metrics path
        kfx.vis.kfp_metrics([
            # render as percent
            kfx.vis.kfp_metric("recall-score", 0.9, percent=true),
            # override metric format with custom value
            kfx.vis.kfp_metric(name="percision-score", value=0.8, metric_format="PERCENTAGE"),
            # render raw score
            kfx.vis.kfp_metric("raw-score", 123.45),
        ]).write_to(mlpipeline_metrics)

        # output visualization metadata to mlpipeline_ui_metadata obj
        kfx.vis.kfp_ui_metadata(
            [
                # creates a confusion matrix vis
                kfx.vis.confusion_matrix(
                    source="gs://your_project/your_bucket/your_cm_file",
                    labels=["True", "False"],
                ),
                # creates a markdown with inline source
                kfx.vis.markdown(
                    "# Inline Markdown: [A link](https://www.kubeflow.org/)",
                    storage="inline",
                ),
                # creates a markdown with a remote source
                kfx.vis.markdown(
                    "gs://your_project/your_bucket/your_markdown_file",
                ),
                # creates a ROC curve with a remote source
                kfx.vis.roc(
                    "gs://your_project/your_bucket/your_roc_file",
                ),
                # creates a Table with a remote source
                kfx.vis.table(
                    "gs://your_project/your_bucket/your_csv_file",
                    header=["col1", "col2"],
                ),
                # creates a tensorboard viewer
                kfx.vis.tensorboard(
                    "gs://your_project/your_bucket/logs/*",
                ),
                # creates a custom web app from a remote html file
                kfx.vis.web_app(
                    "gs://your_project/your_bucket/your_html_file",
                ),
                # creates a Vega-Lite vis as a web app
                kfx.vis.vega_web_app(spec={
                    "$schema": "https://vega.github.io/schema/vega-lite/v4.json",
                    "description": "A simple bar chart with embedded data.",
                    "data": {
                        "values": [
                            {"a": "A", "b": 28}, {"a": "B", "b": 55}, {"a": "C", "b": 43},
                            {"a": "D", "b": 91}, {"a": "E", "b": 81}, {"a": "F", "b": 53},
                            {"a": "G", "b": 19}, {"a": "H", "b": 87}, {"a": "I", "b": 52}
                        ]
                    },
                    "mark": "bar",
                    "encoding": {
                        "x": {"field": "a", "type": "ordinal"},
                        "y": {"field": "b", "type": "quantitative"}
                    }
                })
            ]
        ).write_to(mlpipeline_ui_metadata)


"""
from kfx.vis._helpers import (
    roc,
    table,
    asdict,
    asjson,
    web_app,
    markdown,
    kfp_metric,
    kfp_metrics,
    tensorboard,
    tolocalfile,
    kfp_ui_metadata,
    confusion_matrix,
)



"""Functions to help generate Vega or Vega-Lite spec as web-app in kubeflow pipeline UI."""
import json
import urllib.parse

from typing import Any

import kfx.dsl
import kfx.vis.models

from kfx.vis._helpers import web_app


def _vega_embed_html(
    spec: dict,
    opts: dict = None,
    title="Generated by kfx.vis",
    vega: int = 5,
    vega_lite: int = 4,
) -> str:
    """Returns a html that generates a Vega or Vega-Lite visualization.

    Args:
        spec (dict): Vega or Vega-Lite spec as a dict.
        opts (dict, optional): Options to pass to vega-embed. Defaults to None.
        title (str, optional): Title for the web app. Defaults to "Generated by kfx.vis".
        vega (int, optional): Major version of Vega to use. Defaults to 5.
        vega_lite (int, optional): Major version of Vega-Lite to use. Defaults to 4.
    """

    return f"""
<!DOCTYPE html>
<html>
<head>
  <script src="https://cdn.jsdelivr.net/npm/vega@{vega}"></script>
  <script src="https://cdn.jsdelivr.net/npm/vega-lite@{vega_lite}"></script>
  <!-- Import vega-embed -->
  <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
  <title>{title}</title>
  <style>
    html, body {{
      width: 100%;
      height: 100%;
      margin: 0px;
      border: 0;
    }}
    #vis {{
      width: 100%;
      height: 100%;
    }}
    </style>
</head>
<body>

<div id="vis"></div>

<script type="text/javascript">
  var spec = {json.dumps(spec)};
  var opts = {json.dumps(opts or {})}
  vegaEmbed('#vis', spec, opts).then(function(result) {{
    console.log("Generated with kfx.vis (https://github.com/e2fyi/kfx)!")
  }}).catch(console.error);
</script>
</body>
</html>
"""


def _kfp_ui_api(kfp_artifact: kfx.dsl.KfpArtifact) -> str:
    """Returns the path to call to retrieve the artifact.

    Args:
        kfp_artifact ([kfx.dsl.KfpArtifact]): kubflow flow pipeline artifact.

    Returns:
        str: path to the api to get the artifact.
    """
    return f"artifacts/get?source={kfp_artifact.storage}&bucket={urllib.parse.quote_plus(kfp_artifact.bucket)}&key={urllib.parse.quote_plus(kfp_artifact.key)}"


def _kfp_artifact_to_api(obj: Any) -> Any:
    """Converts any KfpArtifact in a dict into a url to kfp UI api call."""

    if isinstance(obj, kfx.dsl.KfpArtifact):
        return _kfp_ui_api(obj)

    if isinstance(obj, dict):
        return {
            key: _kfp_artifact_to_api(value)
            if isinstance(value, kfx.dsl.KfpArtifact)
            else value
            for key, value in obj.items()
        }

    if isinstance(obj, list):
        return [_kfp_artifact_to_api(item) for item in obj]

    return obj


def vega_web_app(
    spec: dict,
    opts: dict = None,
    title="Generated by kfx.vis",
    vega: int = 5,
    vega_lite: int = 4,
) -> kfx.vis.models.WebApp:
    """Provides the metadata needed for kubeflow pipeline UI to render a `Vega <https://vega.github.io/>`_ or `Vega-Lite <https://vega.github.io/vega-lite/>`_ vis in as a custom web app.

    This web app uses `vega embed <https://github.com/vega/vega-embed>`_ to render the vis.

    Args:
        spec (dict): Vega or Vega-Lite spec as a dict.
        opts (dict, optional): Options to pass to vega-embed. Defaults to None.
        title (str, optional): Title for the web app. Defaults to "Generated by kfx.vis".
        vega (int, optional): Version of Vega to use. Defaults to 5.
        vega_lite (int, optional): Version of Vega-Lite to use. Defaults to 4.

    Returns:
        kfx.vis.models.WebApp: pydantic data object describing a Vega/Vega-Lite web app.
    """
    # so that credential cookies will also be sent
    opts = opts or {}
    opts["loader"] = opts.get("loader", {})
    opts["loader"]["http"] = opts["loader"].get("http", {})
    opts["loader"]["http"].update({"credentials": "same-origin"})

    data = spec.get("data")
    if data:
        # converts any KfpArtifact into api call url
        spec["data"] = _kfp_artifact_to_api(data)

    return web_app(
        source=_vega_embed_html(spec, opts, title, vega=vega, vega_lite=vega_lite),
        storage="inline",
    )



"""Utils."""
import os
import os.path

from typing import Callable, NamedTuple

import kfp.dsl

from kubernetes import client as k8s_client

from kfx.dsl._compat import sanitize_k8s_name

DEFAULT_KEY_FORMAT = "{{workflow.name}}/{{pod.name}}"


class WorkflowVars(NamedTuple):
    """Describes a templated workflow environment variable."""

    name: str
    template: str


def set_workflow_env(
    workflow_vars: WorkflowVars = WorkflowVars(
        name="WORKFLOW_DEFAULT_KEY_FORMAT", template="{{workflow.name}}/{{pod.name}}"
    )
) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
    """Modifier for kubeflow pipelines tasks.

    Setup a kfp op to pass in the workflow variables as an environment var.

    See https://github.com/argoproj/argo/blob/master/docs/variables.md

    Example::

        import kfp
        import kfx.lib.utils as kfxutils

        from kfp.components import func_to_container_op

        @func_to_container_op
        def echo_workflow_vars():
            import os

            print(os.environ.get("WORKFLOW_NAME"))


        @kfp.dsl.pipeline()
        def simple_pipeline():
            op = echo_workflow_vars()

            # op will have an environment variable "WORKFLOW_NAME"
            # that provides the name of the workflow
            op.apply(kfxutils.set_workflow_env(
                kfxutils.WorkflowVars("WORKFLOW_NAME", template="{{workflow.name}}")
            ))

    Args:
        name (str, optional): Env var name to pass in.
            Defaults to "WORKFLOW_DEFAULT_KEY_FORMAT".
        value (str, optional): Template string with argo workflow variables.
            Defaults to "{{workflow.name}}/{{pod.name}}".

    Returns:
        Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: kfp op.
    """

    def apply_workflow_name_env(task: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
        task.container.add_env_variable(
            k8s_client.V1EnvVar(name=workflow_vars.name, value=workflow_vars.template)
        )
        return task

    return apply_workflow_name_env


def set_pod_metadata_envs(
    pod_name: str = "POD_NAME",
    namespace: str = "NAMESPACE",
    node_name: str = "NODE_NAME",
) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
    """Modifier for kubeflow pipelines tasks.

    Setup a kfp op to pass in the pod name, namespace, and node name as env
    var.

    Example::

        import kfp
        import kfx.lib.utils as kfxutils

        from kfp.components import func_to_container_op

        @func_to_container_op
        def echo_podname():
            import os

            print(os.environ.get("POD_NAME"))


        @kfp.dsl.pipeline()
        def simple_pipeline():
            op = echo_podname()
            op.apply(kfxutils.set_pod_metadata_envs())

    Args:
        pod_name (str, optional): Env var name for the pod name.
            Defaults to "POD_NAME".
        namespace (str, optional): Env var name for the namespace.
            Defaults to "NAMESPACE".
        node_name (str, optional): Env var name for the node name.
            Defaults to "NODE_NAME".

    Returns:
        Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: kfp op.
    """

    def apply_pod_metadata_envs(task: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
        for name, field_path in [
            (pod_name, "metadata.name"),
            (namespace, "metadata.namespace"),
            (node_name, "spec.nodeName"),
        ]:

            task.container.add_env_variable(
                k8s_client.V1EnvVar(
                    name=name,
                    value_from=k8s_client.V1EnvVarSource(
                        field_ref=k8s_client.V1ObjectFieldSelector(
                            field_path=field_path
                        )
                    ),
                )
            )
        return task

    return apply_pod_metadata_envs


class ArtifactLocationHelper:
    """Helper class to generate artifact location based on provided argo config.

    See an example of an `Argo configmap <https://github.com/argoproj/argo/blob/master/docs/workflow-controller-configmap.yaml>`_.

    ::

        import kfp.components
        import kfp.dsl
        import kfx.dsl


        helper = kfx.dsl.ArtifactLocationHelper(
            scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
        )

        @kfp.components.func_to_container_op
        def test_op(
            mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
        ):
            "A test kubeflow pipeline task."

            import kfx.dsl
            import kfx.vis

            # write the markdown to the `markdown-data` artifact
            markdown_data_file.write("### hello world")

            # creates an ui metadata object
            ui_metadata = kfx.vis.kfp_ui_metadata(
                # Describes the vis to generate in the kubeflow pipeline UI.
                # In this case, a markdown vis from a markdown artifact.
                [kfx.vis.markdown(kfx.dsl.KfpArtifact("markdown_data_file"))]
                # `KfpArtifact` provides the reference to data artifact created
                # inside this task
            )

            # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
            mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

            # prints the uri to the markdown artifact
            print(ui_metadata.outputs[0].source)


        @kfp.dsl.pipeline()
        def test_pipeline():
            "A test kubeflow pipeline"

            op: kfp.dsl.ContainerOp = test_op()

            # modify kfp operator with artifact location metadata through env vars
            op.apply(helper.set_envs())
    """

    artifact_prefix_env: str = "WORKFLOW_ARTIFACT_PREFIX"
    artifact_storage_env: str = "WORKFLOW_ARTIFACT_STORAGE"
    artifact_bucket_env: str = "WORKFLOW_ARTIFACT_BUCKET"
    artifact_key_prefix_env: str = "WORKFLOW_ARTIFACT_KEY_PREFIX"

    def __init__(
        self, scheme: str, bucket: str, key_prefix: str = "", key_format: str = ""
    ):
        """Creates a new instance of ArtifactLocationHelper object.

        Args:
            scheme (str): Storage scheme, e.g. s3, minio, gcs.
            bucket (str): Name of the bucket.
            key_prefix (str, optional): Key prefix set inside argo artifactory
                configmap. Will not be used if key_format is provided.
                Defaults to "".
            key_format (str, optional): Key format set inside argo artifactory
                configmap. Defaults to "".
        """

        self.scheme = scheme
        self.bucket = bucket
        self.key_prefix = key_prefix
        self.key_format = key_format

    # ref
    # https://github.com/argoproj/argo/blob/f25a45deb4a7179044034da890884432e750d98a/workflow/controller/workflowpod.go#L859
    # artifact location is defaulted using the following formula:
    #   <worflow_name>/<pod_name>/<artifact_name>.tgz
    #   (e.g. myworkflowartifacts/argo-wf-fhljp/argo-wf-fhljp-123291312382/src.tgz)
    def _get_key_prefix(self) -> str:
        """Returns artifact key prefix.

        Returns:
            str: artifact key prefix
        """
        if self.key_format:
            return self.key_format

        return os.path.join(self.key_prefix, DEFAULT_KEY_FORMAT)

    def set_envs(
        self, image: str = "e2fyi/kfx:latest"
    ) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
        """A kfp task modifier.

        This task modifier appends 2 env variables to the task, which
        can be subsequently used to determine the artifact location and prefix.

        Args:
            image (str, optional): image to use for the task. Defaults to
                "e2fyi/kfx:latest".

        Returns:
            Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: modified task.
        """

        def set_workflow_envs(task: kfp.dsl.ContainerOp):
            task.container.image = image
            artifact_prefix = sanitize_k8s_name(task.name)

            for name, value in [
                (self.artifact_storage_env, self.scheme),
                (self.artifact_bucket_env, self.bucket),
                (self.artifact_key_prefix_env, self._get_key_prefix()),
                (self.artifact_prefix_env, artifact_prefix),
            ]:
                task.container.add_env_variable(
                    k8s_client.V1EnvVar(name=name, value=value)
                )

        return set_workflow_envs


def _handle_special_artifact_names(name: str) -> str:
    """Always sanitize special artifact names (e.g. mlpipeline_ui_metadata)"""
    sanitized: str = sanitize_k8s_name(name)
    return (
        sanitized
        if sanitized in {"mlpipeline-ui-metadata", "mlpipeline-metrics"}
        else name
    )


def _sanitize_artifact_name(name: str, sanitize: bool = False) -> str:
    """Sanitize the artifact name based on k8s resource naming convention.

    Also remove suffixes "_path" and "_file". (See this `comment <https://github.com/kubeflow/pipelines/blob/4cb81ea047361ddce7ce8b0b68133b0a92724588/sdk/python/kfp/components/_python_op.py#L327>'_.)


    Args:
        name (str): [description]
        sanitize (bool, optional): Whether to sanitize the name. Defaults to False.

    Returns:
        str: [description]
    """
    if name.endswith("_path"):
        name = name[0 : -len("_path")]
    elif name.endswith("_file"):
        name = name[0 : -len("_file")]

    return (  # type: ignore
        sanitize_k8s_name(name) if sanitize else _handle_special_artifact_names(name)
    )


class KfpArtifact:
    """Class to represent a kubeflow pipeline artifact created inside the pipeline task."""

    def __init__(self, name: str, ext: str = ".tgz", sanitize_name: bool = False):
        """Reference to a kfp artifact that is created within the kubeflow pipeline task.

        This function should be used inside the kfp task. It returns the artifact uri,
        which then can be provided to the kubeflow pipeline UI - i.e. as the
        `source` field inside kubeflow pipeline ui metadata.

        ::

            import kfp.components
            import kfp.dsl
            import kfx.dsl


            helper = kfx.dsl.ArtifactLocationHelper(
                scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
            )

            @kfp.components.func_to_container_op
            def test_op(
                mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
            ):
                "A test kubeflow pipeline task."

                import json

                import kfx.dsl
                import kfx.vis
                import kfx.vis.vega

                data = [
                    {"a": "A", "b": 28},
                    {"a": "B", "b": 55},
                    {"a": "C", "b": 43},
                    {"a": "D", "b": 91},
                    {"a": "E", "b": 81},
                    {"a": "F", "b": 53},
                    {"a": "G", "b": 19},
                    {"a": "H", "b": 87},
                    {"a": "I", "b": 52},
                ]

                # `KfpArtifact` provides the reference to data artifact created
                # inside this task
                spec = {
                    "$schema": "https://vega.github.io/schema/vega-lite/v4.json",
                    "description": "A simple bar chart",
                    "data": {
                        "values": data,
                    },
                    "mark": "bar",
                    "encoding": {
                        "x": {"field": "a", "type": "ordinal"},
                        "y": {"field": "b", "type": "quantitative"},
                    },
                }

                # write the markdown to the `markdown-data` artifact
                markdown_data_file.write("### hello world")

                # creates an ui metadata object
                ui_metadata = kfx.vis.kfp_ui_metadata(
                    # Describes the vis to generate in the kubeflow pipeline UI.
                    [
                        # markdown vis from a markdown artifact.
                        # `KfpArtifact` provides the reference to data artifact created
                        # inside this task
                        kfx.vis.markdown(kfx.dsl.KfpArtifact("markdown_data_file")),
                        # a vega web app from the vega data artifact.
                        kfx.vis.vega.vega_web_app(spec),
                    ]
                )

                # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
                mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

                # prints the uri to the markdown artifact
                print(ui_metadata.outputs[0].source)


            @kfp.dsl.pipeline()
            def test_pipeline():
                "A test kubeflow pipeline"

                op: kfp.dsl.ContainerOp = test_op()

                # modify kfp operator with artifact location metadata through env vars
                op.apply(helper.set_envs())


        Args:
            name (str): name of the artifact.
            ext (str, optional): extension for the artifact. Defaults to ".tgz".
            sanitize_name (bool, optional): whether to sanitize the artifact name. Defaults to False.

        Returns:
            str: uri to the artifact which can be provided to kfp ui.
        """
        self.storage = os.environ[ArtifactLocationHelper.artifact_storage_env]
        self.bucket = os.environ[ArtifactLocationHelper.artifact_bucket_env]
        self.key_prefix = os.environ[ArtifactLocationHelper.artifact_key_prefix_env]
        self.prefix = os.environ[ArtifactLocationHelper.artifact_prefix_env]
        self.name = _sanitize_artifact_name(name, sanitize_name)
        self.ext = ext
        self.key = os.path.join(
            self.key_prefix, "%s-%s%s" % (self.prefix, self.name, self.ext)
        )

    @property
    def source(self) -> str:
        """Url to the artifact source."""
        path = os.path.join(self.bucket, self.key)
        return "%s://%s" % (self.storage, path)

    def __str__(self):
        """Url to the artifact source."""
        return self.source



"""Extension to kfp dsl.

Using `kfx.dsl.ContainerOpTransform` to modify ContainerOp internal properties
::

    import kfp.components
    import kfp.dsl
    import kfx.dsl

    transforms = (
        kfx.dsl.ContainerOpTransform()
        .set_resources(cpu="500m", memory=("1G", "4G"))
        .set_image_pull_policy("Always")
        .set_env_vars({"ENV": "production"})
        .set_env_var_from_secret("AWS_ACCESS_KEY", secret_name="aws", secret_key="access_key")
        .set_annotations({"iam.amazonaws.com/role": "some-arn"})
    )


    @kfp.dsl.components.func_to_container_op
    def echo(text: str) -> str:
        print(text)
        return text


    @kfp.dsl.pipeline(name="demo")
    def pipeline(text: str):
        op1 = echo(text)
        op2 = echo("%s-%s" % text)

        # u can apply the transform on op1 only
        # op1.apply(transforms)

        # or apply on all ops in the pipeline
        kfp.dsl.get_pipeline_conf().add_op_transformer(transforms)


Using `kfx.dsl.ArtifactLocationHelper` to get the path to an artifact generated
by a kfp task.
::

    import kfp.components
    import kfp.dsl
    import kfx.dsl


    helper = kfx.dsl.ArtifactLocationHelper(
        scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
    )

    @kfp.components.func_to_container_op
    def test_op(
        mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
    ):
        "A test kubeflow pipeline task."

        import json

        import kfx.dsl
        import kfx.vis
        import kfx.vis.vega

        data = [
            {"a": "A", "b": 28},
            {"a": "B", "b": 55},
            {"a": "C", "b": 43},
            {"a": "D", "b": 91},
            {"a": "E", "b": 81},
            {"a": "F", "b": 53},
            {"a": "G", "b": 19},
            {"a": "H", "b": 87},
            {"a": "I", "b": 52},
        ]

        # `KfpArtifact` provides the reference to data artifact created
        # inside this task
        spec = {
            "$schema": "https://vega.github.io/schema/vega-lite/v4.json",
            "description": "A simple bar chart",
            "data": {
                "values": data,
            },
            "mark": "bar",
            "encoding": {
                "x": {"field": "a", "type": "ordinal"},
                "y": {"field": "b", "type": "quantitative"},
            },
        }

        # write the markdown to the `markdown-data` artifact
        markdown_data_file.write("### hello world")

        # creates an ui metadata object
        ui_metadata = kfx.vis.kfp_ui_metadata(
            # Describes the vis to generate in the kubeflow pipeline UI.
            [
                # markdown vis from a markdown artifact.
                # `KfpArtifact` provides the reference to data artifact created
                # inside this task
                kfx.vis.markdown(kfx.dsl.KfpArtifact("markdown_data_file")),
                # a vega web app from the vega data artifact.
                kfx.vis.vega.vega_web_app(spec),
            ]
        )

        # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
        mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

        # prints the uri to the markdown artifact
        print(ui_metadata.outputs[0].source)


    @kfp.dsl.pipeline()
    def test_pipeline():
        "A test kubeflow pipeline"

        op: kfp.dsl.ContainerOp = test_op()

        # modify kfp operator with artifact location metadata through env vars
        op.apply(helper.set_envs())

"""
from kfx.dsl._transformers import ContainerOpTransform
from kfx.dsl._artifact_location import (
    KfpArtifact,
    WorkflowVars,
    ArtifactLocationHelper,
    set_workflow_env,
    set_pod_metadata_envs,
)



"""Transform functions that modify containerOp."""
from typing import Dict, List, Tuple, Union, Callable
from fnmatch import fnmatch

import kfp.dsl
import kubernetes.client as k8s

TransformFunc = Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]


class ContainerOpTransform:
    """Helper class to manipulate some common internal properties of ContainerOp.

    Please refer to documentation for full set of transforms available.

    ::

        import kfp.components
        import kfp.dsl
        import kfx.dsl

        transforms = (
            kfx.dsl.ContainerOpTransform()
            .set_resources(cpu="500m", memory=("1G", "4G"))
            .set_image_pull_policy("Always")
            .set_env_vars({"ENV": "production"})
            .set_env_var_from_secret("AWS_ACCESS_KEY", secret_name="aws", secret_key="access_key")
            .set_annotations({"iam.amazonaws.com/role": "some-arn"})
        )


        @kfp.dsl.components.func_to_container_op
        def echo(text: str) -> str:
            print(text)
            return text


        @kfp.dsl.pipeline(name="demo")
        def pipeline(text: str):
            op1 = echo(text)
            op2 = echo("%s-%s" % text)

            # u can apply the transform on op1 only
            # op1.apply(transforms)

            # or apply on all ops in the pipeline
            kfp.dsl.get_pipeline_conf().add_op_transformer(transforms)

    """

    def __init__(self, transforms: List[TransformFunc] = None):
        """Creates a new instance of ContainerOpTransform object.

        Args:
            transforms (List[TransformFunc], optional): Optional list of custom transform functions. Defaults to None.
        """
        self._transforms: List[TransformFunc] = transforms or []

    def __call__(self, op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
        """In-place transform of the provided ContainerOp.

        Args:
            op (kfp.dsl.ContainerOp): ContainerOp obj.

        Returns:
            kfp.dsl.ContainerOp: ContainerOp obj.
        """
        [  # pylint: disable=expression-not-assigned
            transform(op) for transform in self._transforms
        ]
        return op

    def set_annotations(self, annotations: Dict[str, str]) -> "ContainerOpTransform":
        """Update the transform function to set the provided annotations to the ContainerOp.

        Args:
            annotations (Dict[str, str]): dict of annotation keys and values.
        """

        def set_annotations_transform(op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
            [  # pylint: disable=expression-not-assigned
                op.add_pod_annotation(name, value)
                for name, value in annotations.items()
            ]
            return op

        self._transforms.append(set_annotations_transform)
        return self

    def set_labels(self, labels: Dict[str, str]) -> "ContainerOpTransform":
        """Update the transform function to set the provided labels to the ContainerOp.

        Args:
            labels (Dict[str, str]): dict of labels keys and values.
        """

        def set_labels_transform(op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
            [  # pylint: disable=expression-not-assigned
                op.add_pod_label(name, value) for name, value in labels.items()
            ]
            return op

        self._transforms.append(set_labels_transform)
        return self

    def add_env_vars(self, env_vars: Dict[str, str]) -> "ContainerOpTransform":
        """Update the transform function to set the provided env vars to the ContainerOp.

        Args:
            env_vars (Dict[str, str]): dict of env vars keys and values.
        """

        def set_env_vars_transform(op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
            [  # pylint: disable=expression-not-assigned
                op.add_env_variable(k8s.V1EnvVar(name, value))
                for name, value in env_vars.items()
            ]
            return op

        self._transforms.append(set_env_vars_transform)
        return self

    def add_env_var(self, name: str, value: str) -> "ContainerOpTransform":
        """Update the transform function to set the provided env var to the ContainerOp.

        Args:
            name (str): name of the env var.
            value (str): value of the env var.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(
            lambda op: op.container.add_env_variable(k8s.V1EnvVar(name, value))
        )
        return self

    def add_env_var_from_secret(
        self, name: str, secret_name: str, secret_key: str
    ) -> "ContainerOpTransform":
        """Update the transform function to set the provided env var from a k8s secret.

        Args:
            name (str): name of the env var.
            secret_name (str): name of the k8s secret.
            secret_key (str): key to retrieve from the k8s secret.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(
            lambda op: op.container.add_env_variable(
                k8s.V1EnvVar(
                    name,
                    value_from=k8s.V1EnvVarSource(
                        secret_key_ref=k8s.V1SecretKeySelector(
                            key=secret_key, name=secret_name
                        )
                    ),
                )
            )
        )
        return self

    def add_env_var_from_configmap(self, configmap_name: str) -> "ContainerOpTransform":
        """Update the transform function to set env vars from a configmap.

        Args:
            configmap_name (str): name of the configmap.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(
            lambda op: op.container.add_env_from(
                k8s.V1EnvFromSource(
                    config_map_ref=k8s.V1ConfigMapEnvSource(name=configmap_name)
                )
            )
        )
        return self

    def set_cpu_resources(
        self, request: Union[int, str], limit: Union[int, str] = None
    ) -> "ContainerOpTransform":
        """Update the transform function to set the cpu resources for the main container.

        Args:
            request (Union[int, str]): How much cpu to request.
            limit (Union[int, str], optional): Max cpu load before throttling. Defaults to None.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """

        self._transforms.append(
            lambda op: op.container.set_cpu_request(str(request)).set_cpu_limit(
                str(limit or request)
            )
        )

        return self

    def set_memory_resources(
        self, request: Union[int, str], limit: Union[int, str] = None
    ) -> "ContainerOpTransform":
        """Update the transform function to set the memory resources for the main container.

        Args:
            request (Union[int, str]): How much memory to request.
            limit (Union[int, str], optional): Max memory before killing the pod. Defaults to None.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(
            lambda op: op.container.set_memory_request(str(request)).set_memory_limit(
                str(limit or request)
            )
        )

        return self

    def set_gpu_limit(
        self, value: Union[int, str], vendor: str = "nvidia"
    ) -> "ContainerOpTransform":
        """Update the transform function to set the cpu limit for the main container.

        Args:
            value (Union[int, str]): GPU limit for the main container.
            vendor (str, optional): Either "nvidia" or "amd". Defaults to "nvidia".

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(
            lambda op: op.container.set_gpu_limit(str(value), vendor)
        )
        return self

    def set_image_pull_policy(self, policy: str) -> "ContainerOpTransform":
        """Update the transform function to set the image pull policy for the main container.

        Args:
            policy (str): One of "Always", "Never", "IfNotPresent".

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        self._transforms.append(lambda op: op.container.set_image_pull_policy(policy))
        return self

    def set_sidecar_image_pull_policy(
        self, policy: str, sidecar_name: str = "*"
    ) -> "ContainerOpTransform":
        """Update the transform function to set the image pull policy for the sidecars.

        Args:
            policy (str): One of "Always", "Never", "IfNotPresent".
            sidecar_name (str, optional): Glob pattern for sidecar name. Defaults to "*".

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """

        def set_sidecar_transform(op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
            [  # pylint: disable=expression-not-assigned
                sidecar.set_image_pull_policy(policy)
                for sidecar in op.sidecars
                if fnmatch(sidecar.name, sidecar_name)
            ]

        self._transforms.append(set_sidecar_transform)
        return self

    def set_resources(
        self,
        cpu: Union[int, str, Tuple[Union[int, str], Union[int, str]]] = None,
        memory: Union[str, Tuple[str, str]] = None,
    ) -> "ContainerOpTransform":
        """Update the transform function to set the cpu and memory resources for the main container.

        If a int or str is provided, the resource request will equal the limit (i.e. QoS is Guaranteed).
        Otherwise, a tuple should be provided specifying the request and the limit.

        Args:
            cpu (Union[int, str, Tuple[Union[int, str], Union[int, str]]], optional): A str or tuple representing the cpu request and limit.
            memory (Union[str, Tuple[str, str]], optional): A str or tuple representing the memory request and limit.

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        if isinstance(cpu, (tuple, list)):
            cpu_request, cpu_limit = cpu
        else:
            cpu_request = cpu  # type: ignore
            cpu_limit = cpu  # type: ignore

        if isinstance(memory, (tuple, list)):
            memory_request, memory_limit = memory
        else:
            memory_request = memory  # type: ignore
            memory_limit = memory  # type: ignore

        def set_resources_transform(op: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:

            if cpu_request:
                op.container.set_cpu_request(str(cpu_request))
            if cpu_limit:
                op.container.set_cpu_limit(str(cpu_limit))
            if memory_request:
                op.container.set_memory_request(memory_request)
            if memory_limit:
                op.container.set_memory_limit(memory_limit)
            return op

        self._transforms.append(set_resources_transform)
        return self

    def set_sidecar_resources(
        self,
        cpu: Union[int, str, Tuple[Union[int, str], Union[int, str]]] = None,
        memory: Union[str, Tuple[str, str]] = None,
        sidecar_name: str = "*",
    ) -> "ContainerOpTransform":
        """Update the transform function to set the cpu and memory resources for the sidecars.

        If a int or str is provided, the resource request will equal the limit (i.e. QoS is Guaranteed).
        Otherwise, a tuple should be provided specifying the request and the limit.

        Args:
            cpu (Union[int, str, Tuple[Union[int, str], Union[int, str]]], optional): A str or tuple representing the cpu request and limit.
            memory (Union[str, Tuple[str, str]], optional): A str or tuple representing the memory request and limit.
            sidecar_name (str, optional): Glob pattern matching the sidecar name. Defaults to "*".

        Returns:
            ContainerOpTransform: updated ContainerOpTransform object.
        """
        if isinstance(cpu, (tuple, list)):
            cpu_request, cpu_limit = cpu
        else:
            cpu_request = cpu  # type: ignore
            cpu_limit = cpu  # type: ignore

        if isinstance(memory, (tuple, list)):
            memory_request, memory_limit = memory
        else:
            memory_request = memory  # type: ignore
            memory_limit = memory  # type: ignore

        def set_sidecar_resources_transform(
            op: kfp.dsl.ContainerOp
        ) -> kfp.dsl.ContainerOp:
            for sidecar in op.sidecars:
                if fnmatch(sidecar.name, sidecar_name):
                    if cpu_request:
                        sidecar.set_cpu_request(str(cpu_request))
                    if cpu_limit:
                        sidecar.set_cpu_limit(str(cpu_limit))
                    if memory_request:
                        sidecar.set_memory_request(memory_request)
                    if memory_limit:
                        sidecar.set_memory_limit(memory_limit)
            return op

        self._transforms.append(set_sidecar_resources_transform)
        return self
